import os
import imageio

from PyQt5.QtWidgets  import QFileDialog, QErrorMessage, QMessageBox

from pyrateshield.model import Model



def safe_load_model(filename=None):
    # filename only used to extract current folder
    confirm = confirm_changes("Load Project")
    if not confirm:
        return
    
    newfile = select_project_file(filename, must_exist=True)

    if newfile is not None:
        try:
            model = Model.load_from_project_file(newfile)
        except IOError:
            error_dialog = QErrorMessage()
            error_dialog.showMessage(f'Could not read from {filename}!')
            error_dialog.exec_()
            model = None
            raise
    else:
        model = None
        
    return model

def show_help(title="title", text="Help?"):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Question)
    msg.setText("Help")
    msg.setInformativeText(text)
    msg.setStandardButtons(QMessageBox.Ok)
    msg.setWindowTitle(title)
    msg.exec_()
    
    
    

def safe_write_model(model=None, filename=None):
    if filename is None:
        filename = select_project_file(model.filename)
        
    if filename is not None:
        
        try:
            model.save_to_project_file(filename)
            model.filename = filename
        except:
            error_dialog = QErrorMessage()
            error_dialog.showMessage(f'Could not write to {filename}!')
            error_dialog.exec_()
            raise
    

def safe_load_image():

    
    filename = ask_image_file('Open bitmap image')
    image = None
    if filename:
        try:
            with open(filename, 'rb') as f:                
                image = imageio.imread(filename)
        except:
            error_dialog = QErrorMessage()
            error_dialog.showMessage(f'Could not read from {filename}!')
            error_dialog.exec_()
            

    return image
        
       
def ask_image_file(title):
    extensions = "Image Files (*.png *.jpg *.jpeg *.tiff *.bmp)"
    file = ask_existing_file(title, extensions)
    if file == '':
        file = None
    return file

def ask_new_image_file(title):
    extensions = "Image Files (*.png *.jpg *.jpeg *.tiff *.bmp)"
    file = ask_new_file(title, extensions)
    if file == '':
        file = None
    return file

def ask_existing_file(title, extensions, directory=None):
    filedialog = QFileDialog(directory=directory)
    filedialog.setFileMode(QFileDialog.ExistingFile)
    file = str(QFileDialog.getOpenFileName(filedialog, title, "",
                                           extensions)[0])
    if file == '': # Cancel
        file = None
    return file

def ask_new_file(title, extensions, directory=None):
    filedialog = QFileDialog(directory=directory)
    filedialog.setFileMode(QFileDialog.AnyFile)
    file = str(QFileDialog.getSaveFileName(filedialog, title, "",
                                           extensions)[0])
    
    if file == '': # Cancel
        file = None
    
    return file
           
def select_project_file(filename=None, must_exist=False):
    # filename only used for finding current folder
    
    if filename is not None:
        folder = os.path.split(filename)[0]
        if not os.path.exists(folder):
            folder = None
    else:
        folder = None
        
    if must_exist:
        file_selector = ask_existing_file
    else:
        file_selector = ask_new_file
        
        
    project_file = file_selector("Select Project File", 
                                 "PyrateShield Projects (*.zip *.psp)",
                                 directory=folder)
    
    return project_file

def confirm_new_project():
    confirm = confirm_changes("New Project")
    if confirm:
        return True
    else:
        return False

def pixelsize_error(title=None):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText("Error!")
    txt = 'Real world distance in cm must be positive and greater than 0!'
    msg.setInformativeText(txt)
    msg.setStandardButtons(QMessageBox.Ok)
    msg.setWindowTitle(title)
    msg.exec_()
    

def confirm_changes(title=None):
   msg = QMessageBox()
   msg.setIcon(QMessageBox.Warning)
   msg.setText("Warning!")
   msg.setInformativeText("Any unsaved changes will be lost!")
   msg.setWindowTitle(title)
   msg.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
   
  
   confirm = msg.exec_()
   if confirm == QMessageBox.Ok:
       return True
   else:
       return False
  
        
    

    

        


      
        



from PyQt5.QtCore import QObject, pyqtSignal
import inspect
import yaml
from pyrateshield import labels
import pyrateshield

#LOG_LEVEL = 10

def equal_position(position1, position2):
    if  [position1, position2].count(None) == 1:
        return False
    elif [position1, position2].count(None) == 2:
        return True
    else:
        return position1[0] == position2[0]\
            and position1[1] == position2[1]
        


def equal_materials(materials1, materials2):
    if  [materials1, materials2].count(None) == 1:
        return False
    elif [materials1, materials2].count(None) == 2:
        return True
    elif len(materials1) != len(materials2):
            return False
    else:
        eq = True
        for material1, material2 in zip(materials1, materials2):
            eq = eq and material1[0] == material2[0] and material1[1] == material2[1]
        return eq
    

### Base class
class ModelItem(QObject):
    label = 'ModelItem'
    
    update_event = pyqtSignal(object)
    
    
    _attr_dct = {}
    _attr_defaults = {}   


    def __init__(self, **kwargs):        
        
        QObject.__init__(self)

        
        init_values = self._get_default_values()
        init_values.update(kwargs)
        
        for attr_name, value in init_values.items():
        
            if attr_name in self._attr_dct.keys():
                # set by attribute name
                setattr(self, attr_name, value)
            elif attr_name in self._attr_dct.values():
                # set by label
                setattr(self, self.attr_name_from_label(attr_name), value)
            else:
                pass
                #raise AttributeError(f'Cannot set {attr_name} for class {self.__class__.__name__}')
                            
    def get_attr_from_label(self, label):
        return getattr(self, self.attr_name_from_label(label))
        
    def delete(self):
        self.delete.emit(self)
        

    def __copy__(self):
        return self.__class__.from_dict(self.to_dict())
    
    # Code to make objects pickable
    def __getstate__(self):
        return self.to_dict()
    
    def __setstate__(self, dct):
        self.__init__(**dct)
    # end pickable

    @classmethod
    def _get_default_value(cls, attr_name):
        if cls._attr_dct[attr_name] in cls._attr_defaults.keys():     
            # default value defined
            value = cls._attr_defaults[cls._attr_dct[attr_name]]

        elif attr_name in cls._attr_dct.keys(): 
            # No default value but is in the attribute dct
            value = None
        else:
            # Debug purpuses should not happen for a well defined ModelItem
            raise AttributeError(f'Unknown Attribute {attr_name}')
            
        if inspect.isclass(value) or inspect.isfunction(value):
            value = value() # call or create instance
        return value
    
    @classmethod
    def _get_default_values(self):
        values = {}
        for attr_name in self._attr_dct.keys():
            values[attr_name] = self._get_default_value(attr_name)
        return values
        
    def __setattr__(self, attr_name, value):

        if attr_name not in self._attr_dct.keys():
            return super().__setattr__(attr_name, value)
        

        if hasattr(self, attr_name):
            old_value = getattr(self, attr_name)
        else: 
            super().__setattr__(attr_name, None)
            old_value = None
            
        if hasattr(self, attr_name):
            pass
            #print(f'!! set attr {attr_name} to {str(value)} from {str(old_value)} with type {type(value)}')
       
        
        # Override None values to default values                
        if value is None:            
            value = self._get_default_value(attr_name)
           
        
        if attr_name == 'materials' and isinstance(value, list)\
            and equal_materials(old_value, value):            
            return # don't set and don't emit event
                
        elif attr_name == 'position':
            value = [float(value[0]), float(value[1])] # force list!
            if equal_position(old_value, value):
                return # don't set and don't emit event
                
        elif attr_name in ('vertices', 'vertices_pixels') :# force list!    
            
            if not pyrateshield.model_items.Wall.valid_vertices(value):
                msg =f'Invalid value for vertices, {type(value)} and value {value}'
                raise ValueError(msg)
            
                
            if pyrateshield.model_items.Wall.equal_vertices(old_value, value):
                return # don't set and don't emit event
            
                
        elif attr_name == 'image': 
            if old_value is value:
                return
        else:
            if old_value == value:
                return  # don't set and don't emit event

        super().__setattr__(attr_name, value)
        
        if isinstance(value, (ModelItem, 
                      pyrateshield.model_sequences.ModelItemSequence)):
            value.setParent(self)
            
        if attr_name in self._attr_dct.keys():
            # gather event_data and emit event
            value = getattr(self, attr_name)
            label = self._attr_dct[attr_name]
            event_data = self, label, old_value, value
            
           
            self.update_event.emit(event_data)
      
            
    def update_by_dict(self, dct):
        for key, value in dct.items():
            attr_name = self.attr_name_from_label(key)
            setattr(self, attr_name, value)
          
    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)
    
    def enable(self, value):
        if labels.ENABLED in self._attr_dct.values():
            self.enabled = value
        else:
            raise AttributeError

    def to_dict(self):
        dct = {label: getattr(self, var)\
               for var, label in self._attr_dct.items()}
        return dct
    
    @classmethod
    def attr_name_from_label(cls, label): 
        if label in cls._attr_dct.keys():
            return label
        else:
            index = list(cls._attr_dct.values()).index(label)
            return list(cls._attr_dct.keys())[index]
    
    
    def __str__(self):
        #return "\n".join([f"{k}: {v}" for k, v in self.to_dict().items()])
        return yaml.dump(self.to_dict())
    
    def __repr__(self):
        return self.__str__()
    
    def tooltip(self):
        return str(self)
    
    




import os
import yaml
from pyrateshield import labels
from pyrateshield.modelitem import ModelItem



CONSTANTS_FILE = "constants.yml"
ISOTOPES_FILE = "isotopes.yml"

class Isotope(ModelItem):
    _attr_dct = {
        "name": labels.NAME,
        "half_life": labels.HALF_LIFE,
        "self_shielding_options": labels.SELF_SHIELDING_OPTIONS,
    }
    spectrum = None
    
    def __eq__(self, other):
        return self.name == other.name
    
    def mass_number(self):
        mass_number = self.name.split('-')[1]
        try:
            mass_number = int(mass_number)
        except ValueError:
            # Cut off the m for Tc-99m
            mass_number = int(mass_number[:-1])
        return mass_number
    
    def __gt__(self, other):
        return self.mass_number() > other.mass_number()
    
    def __lt__(self, other):
        return self.mass_number() < other.mass_number()
    
    def __ge__(self, other):
        return self.mass_number() >= other.mass_number()
    
    def __le__(self, other):
        return self.mass_number() <= other.mass_number()
        

class ArcherParamsCT(ModelItem):
    _attr_dct = {
        "kvp": labels.KVP,
        "archer": labels.ARCHER,
        "scatter_fraction_body": labels.CT_SCATTER_FRACTION_BODY,
        "scatter_fraction_head": labels.CT_SCATTER_FRACTION_HEAD,
    }

class ArcherParamsXray(ModelItem):
    _attr_dct = {
        "kvp": labels.KVP,
        "archer": labels.ARCHER,
        "scatter_fraction": labels.XRAY_SCATTER_FRACTION,
    }



class Constants:
    _isotopes = None
    _ct = None
    _xray = None
    _wdir = None
    _constants_yml = None
    _isotope_yml = None
   
    def __init__(self):
        self.base_materials = [labels.EMPTY_TABLE] + self.constants_yml[labels.BASE_MATERIALS]
        self.buildup_materials = [labels.EMPTY_TABLE] + self.constants_yml[labels.BUILDUP_MATERIALS]
        self.CT_body_part_options = self.constants_yml[labels.CT_BODY_PART_OPTIONS]
        self.self_shielding_pyshield = self.constants_yml[labels.SELF_SHIELDING_PYSHIELD]
        self.self_shielding_options = [labels.SELF_SHIELDING_NONE,
                                       labels.SELF_SHIELDING_BODY, 
                                       labels.SELF_SHIELDING_FACTOR]
     
        self.decay_chains = self.isotope_yml[labels.DECAY_CHAINS]
        
        supported_isotopes = self.isotope_yml[labels.SUPPORTED_ISOTOPES]
        self.radtracer_supported_isotopes = supported_isotopes[labels.RADTRACER]
        self.pyshield_supported_isotopes = supported_isotopes[labels.PYSHIELD]
        self.isotope_spectra = self.isotope_yml[labels.ISOTOPE_SPECTRA]
        
    @property
    def wdir(self):
        if self._wdir is None:
            try:
                wdir = os.path.split(__file__)[0] 
            except:
                wdir = os.getcwd()
            self._wdir = wdir
        return self._wdir
    
    @property
    def constants_yml(self):
        if self._constants_yml is None:
            with open(os.path.join(self.wdir, CONSTANTS_FILE)) as f:
                constants_yml = yaml.safe_load(f)
            self._constants_yml = constants_yml
        return self._constants_yml
    
    @property
    def isotope_yml(self):
        if self._isotope_yml is None:
            with open(os.path.join(self.wdir, ISOTOPES_FILE)) as f:
               self._isotope_yml = (yaml.safe_load(f))    
        return self._isotope_yml
            
    @property
    def ct(self):
        if self._ct is None:
            self._ct = [ArcherParamsCT.from_dict(item)\
                        for item in self.constants_yml[labels.CT_PARAMETERS]]
        return self._ct
    @property
    def xray(self):
        if self._xray is None:
            self._xray = [ArcherParamsXray.from_dict(item)\
                          for item in self.constants_yml[labels.XRAY_PARAMETERS]]
        return self._xray
    
    
   
    
    @property
    def isotopes(self):
        if self._isotopes is None:

            isotopes = [Isotope.from_dict(item)\
                        for item in self.isotope_yml[labels.ISOTOPES]]
                
            # sort isotopes by mass number
            isotopes = sorted(isotopes)
        
            for isotope in isotopes:
                spectrum = list(self.isotope_spectra[isotope.name])
                spectrum_with_parent = [(isotope.name, energy, intensity)\
                                        for energy, intensity in spectrum]
                for daughter, abundance in self.decay_chains.get(isotope.name, []):
                    for energy, intensity in self.isotope_spectra[daughter]:
                        spectrum.append([energy, intensity*abundance])
                        spectrum_with_parent.append([daughter, energy, intensity*abundance])
                isotope.spectrum = spectrum
                isotope.spectrum_with_parent = spectrum_with_parent
            self._isotopes = isotopes
        return self._isotopes
    
    def get_isotope_name_list(self):
        return [item.name for item in self.isotopes]
        

    def get_isotope_by_name(self, name):
       
       isotope = [x for x in self.isotopes if x.name == name]
       if len(isotope) == 0:
           raise KeyError(f'No isotope with name {name}')
       return isotope[0]
    
  
    
CONSTANTS = Constants()



"""
Generate dosemap, based on pre-simulated (MCNP6) transmission probabilities

Rob van Rooij
"""

import numpy as np
from skimage.draw import polygon
from scipy import interpolate
import os
import pickle

from pyrateshield.shared import time_integrated_activity_coefficient_mbqh


_folder = os.path.dirname(os.path.abspath(__file__))
MCNP_LOOKUP = pickle.load(open(os.path.join(_folder, "MCNP.pickle"), "rb"))

DEFAULT_DENSITY = {
    "Water":           0.998207,
    "Lead":            11.35,
    "Concrete":        2.30,
    "Concrete-Barite": 3.35,
    "Gypsum":          1.0,
    "Brick":           1.8,
}


class TransmissionMCNP:
    def __init__(self, LUT):
        self.LUT = LUT
        self.interp = {}
        
    def get(self, material, thickness):
        if material not in self.LUT:
            return 1
        
        if material not in self.interp:
            x, y = self.LUT[material].T
            self.interp[material] = interpolate.interp1d(x, np.log(y), fill_value="extrapolate")
        
        return np.exp(self.interp[material](thickness))


class TransmissionArcher:
    _ALTERNATIVES = {
        "Water":  ("Concrete", 2.35, 1.0), # This makes little sense, but neither does a wall of water ;-)
        "Concrete-Barite":  ("Concrete", 2.35, 3.35),
        "Gypsum": ("Concrete", 2.35, 0.8),
        "Brick":  ("Concrete", 2.35, 1.6),
    }
    def __init__(self, archer_params):
        self.archer_params = archer_params
        
    def get(self, material, thickness):
        if material in self.archer_params:
            a, b, g = self.archer_params[material]
            
        elif material in self._ALTERNATIVES:
            alt = self._ALTERNATIVES[material]            
            print(f"No transmission parameters for {material} found. Using {alt[0]} with density {alt[2]}")
            a, b, g = self.archer_params[alt[0]]
            a *= alt[2]/alt[1]
            b *= alt[2]/alt[1]
        else:
            print(f"No transmission parameters for {material} found. Ignoring this shielding")
            return 1
            
        return ( (1 + b/a)*np.exp(a*g*thickness) - b/a )**(-1/g) 
        

def source_transmission(source, project):
    source_type = source.__class__.__name__
    
    if source_type == "SourceNM":
        name = source.isotope
        
        self_shielding_key = source.self_shielding if isinstance(source.self_shielding, str) else 'None'
        transmission = TransmissionMCNP(MCNP_LOOKUP[source.isotope][self_shielding_key])
        h_10 = MCNP_LOOKUP[source.isotope][self_shielding_key]["h(10) [uSv/h per MBq/m^2]"]
        
        clearance_model = project.clearances.itemByName(source.clearance)        
        
        source_dose = h_10 * time_integrated_activity_coefficient_mbqh(source, clearance_model=clearance_model)
        
        # Apply numeric self shielding factor here
        if isinstance(source.self_shielding, float):
            source_dose *= source.self_shielding
        
    elif source_type == "SourceCT":
        params = [x for x in project.constants.ct if x.kvp == source.kvp][0]
        transmission = TransmissionArcher(params.archer)
        scatter_frac = {
            "Body": params.scatter_fraction_body, 
            "Head": params.scatter_fraction_head}[source.body_part]
        source_dose = source.number_of_exams * source.dlp * scatter_frac
        
    elif source_type == "SourceXray":
        params = [x for x in project.constants.xray if x.kvp == source.kvp][0]
        transmission = TransmissionArcher(params.archer)
        source_dose = source.number_of_exams * source.dap * params.scatter_fraction
    else:
        raise ValueError("Source type unknown: {}".format(modality))

    return transmission, source_dose * 1E-3 #1E-3 to convert uSv to mSv


def unit_vector(vector):
    return vector / np.linalg.norm(vector)


def get_polygon(imshape, source, line):
    p1 = np.array(line[0])
    p2 = np.array(line[1])
    
    uv1 = unit_vector(p1-source)
    uv2 = unit_vector(p2-source)
    angle = np.arccos(np.clip(np.dot(uv1, uv2), -1.0, 1.0))
    
    min_dist = np.sqrt(imshape[0]**2 + imshape[1]**2)+1
    d = min_dist/np.cos(0.5*angle)

    p3 = source + d*uv1
    p4 = source + d*uv2

    poly = np.array((p1, p3, p4, p2)).T
    
    return polygon(poly[0], poly[1], imshape)


def get_diagonal_intersection_factor(poly, source_ji, vertices_ji):
    """ 
    Get diagonal distance factor through wall for every point within polygon (head-on=1, 45deg=sqrt(2))
    """
    r = (poly[0]-source_ji[0], poly[1]-source_ji[1]) # Vectors from source to points in polygon
    w = (vertices_ji[1][1]-vertices_ji[0][1], -vertices_ji[1][0]+vertices_ji[0][0]) # Vector orthogonal to wall (vec=v1-v0, w=(vec_y,-vec_x))
    r_in_w = r[0]*w[0] + r[1]*w[1] # Inner product for every vector in r with w
    mag_r_w = np.linalg.norm(r, axis=0) * (w[0]**2 + w[1]**2)**0.5 # Multiplied magnitude of vectors
    return np.abs(mag_r_w / r_in_w) # 1 / Cos(th)     ( cos(th) = u.v / (|u| |v|) )


def grid_coords(coords_cm, extent, gridshape):
    x, y = coords_cm
    x0, x1, y0, y1 = extent
    j = (y1-y)/(y1-y0) * gridshape[0] - 0.5
    i = (x-x0)/(x1-x0) * gridshape[1] - 0.5
    return np.array((j, i))

def ccw(A,B,C):
    # Is counter-clockwise?
    return (C[1]-A[1]) * (B[0]-A[0]) > (B[1]-A[1]) * (C[0]-A[0])

def intersect(A,B,C,D):
    # Return true if line segments AB and CD intersect
    return ccw(A,C,D) != ccw(B,C,D) and ccw(A,B,C) != ccw(A,B,D)


def get_walls_by_material(project):
    wall_dict = {}    
    shielding_dict = {s.name: s for s in project.shieldings}
    for wall in project.walls.simplifiedWalls():
        shielding = shielding_dict[wall.shielding]
        
        for material_name, thickness in shielding.materials:
            material_item = project.materials.itemByName(material_name)
        
            material = material_item.radtracer_material
            density = material_item.density
            
            #Allow user-defined densities by modifying the wall thickness instead
            default_density = DEFAULT_DENSITY.get(material, 1)            
            thickness *= density/default_density
            
            entry = wall_dict.setdefault(material, [])
            entry.append( (wall.vertices, thickness) )
    return wall_dict


def pointdose_single_source(point, source, project):        
    transmission, source_dose = source_transmission(source, project)    
    total_transmission = 1
    wall_dict = get_walls_by_material(project)
    for material, wall_list in wall_dict.items():        
        cumul_thickness = 0
        for vertices, thickness in wall_list:        
            if not intersect(point, source.position, vertices[0], vertices[1]):
                continue
            diag_factor = get_diagonal_intersection_factor(point, source.position, vertices)
            cumul_thickness += thickness*diag_factor
        
        total_transmission *= transmission.get(material, cumul_thickness)

    r_squared = (point[0] - source.position[0])**2 + \
                (point[1] - source.position[1])**2
    return source_dose * total_transmission * 100**2 / r_squared


def dosemap_single_source(source, project):
    transmission, source_dose = source_transmission(source, project)
    source_ji = project.dosemap.to_grid_coords(source.position)
    transmission_map = np.ones(project.dosemap.shape)
    wall_dict = get_walls_by_material(project)
    for material, wall_list in wall_dict.items():        
        cumul_thickness_map = np.zeros(project.dosemap.shape)
        for vertices, thickness in wall_list:
            vertices_ji = [project.dosemap.to_grid_coords(vert) for vert in vertices]            
            poly = get_polygon(cumul_thickness_map.shape, source_ji, vertices_ji)
            diag_factors = get_diagonal_intersection_factor(poly, source_ji, vertices_ji)
            cumul_thickness_map[poly] += thickness*diag_factors

        if 0: # For debugging: show the cumulative thickness maps per source per material
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots()
            ax.imshow(cumul_thickness_map, cmap="jet")
            fig.suptitle(f"{source.name}: {material}")
            plt.show()

        transmission_map *= transmission.get(material, cumul_thickness_map)
        
    r_squared = (project.dosemap.grid.X - source.position[0])**2 + \
                (project.dosemap.grid.Y - source.position[1])**2
    return source_dose * transmission_map * 100**2 / r_squared





import pyperclip
import yaml
import numpy as np
import qtawesome as qta

from copy import deepcopy


from PyQt5.QtWidgets import QMenu, QAction
from PyQt5.QtGui import QCursor
from PyQt5.QtCore import QPoint
from pyrateshield import labels
from pyrateshield.gui.graphics_items import (FloorplanPixmap, 
                                             GraphicsModelItem, WallMarker, 
                                             GraphicsPoint, Cross)
from functools import partial

from pyrateshield.model_items import Wall
#from pyrateshield.gui.mpl_items_controller import PixelSizeLine

SNAP_RADIUS_CM = 100



class ContextMenuController():
   
    
    def __init__(self, model=None, view=None, parent=None):
        
        self.parent = parent
        self.view = view
        self.set_model(model)
        
        self.view.rightMouseClick.connect(self.right_click)
        

    def set_model(self, model):
        self.model = model
   
    def right_click(self, event, pixel_item):
        self.event = event
        pos = self.view.mapToScene(event.pos())
        
        if isinstance(pixel_item, FloorplanPixmap):
            self.show_canvas_context_menu(pos)
            event.accept()
    
        elif isinstance(pixel_item, (GraphicsModelItem, WallMarker))\
                        and pixel_item.showContextMenu:            
            self.show_pick_context_menu(pos, pixel_item)
            event.accept()
       
            
            
        
            
    @property
    def copied_items(self):
        # get plain text from clipboard
        text = pyperclip.paste()
        
        try: 
            items = yaml.safe_load(text)
        except:
            print('Invalid yaml on clipboard!')
            return None
        
        if not isinstance(items, dict):
            print('No data dictionary could be loaded from clipboard')
            return None
        
        for key in items.keys():
            if key not in [labels.WALLS, labels.SOURCES_CT, 
                           labels.SOURCES_NM, labels.SOURCES_XRAY,
                           labels.CRITICAL_POINTS]:
                print(f'Invalid key found in dictionary: {key}')
                return None
        copied_items = {}
        for key, dict_items in items.items():
            if not isinstance(dict_items, list):
                print(f'{key} does not contain a list!')
                return None
            copied_items[key] = []            
            item_class = self.model.item_class_for_label(key)
            for dict_item in dict_items:
                try:
                    item = item_class.from_dict(dict_item)
                except:
                    print('Could not convert to {item_class.__class__.__name__}:\n {yaml.dump(dict_item)}')
                    return None
                
                if key == labels.SOURCES_NM:
                    if item.clearance not in self.model.clearances.itemNames():
                        print(f'Clearance {item.clearance} not defined!')
                elif key == labels.WALLS:
                    if item.shielding not in self.model.shieldings.itemNames():
                        print(f'Shielding {item.shielding} not definied!')
                
                copied_items[key] += [item]
        return copied_items
            
                
      
        
    
    def show_canvas_context_menu(self, pos=None, _=None):
       
        
        context_menu = QMenu(self.view)
        pos_cm = self.model.floorplan.geometry.pixels_to_cm([pos.x(), pos.y()])
        
        toolbar = self.parent.toolbar_controller
        for label in [labels.SOURCES_CT, labels.SOURCES_XRAY,
                      labels.SOURCES_NM, labels.CRITICAL_POINTS, labels.WALLS]:
            
      
            title = 'Add ' + label + ' item'
            action = context_menu.addAction(title)
        
            interactive = label == labels.WALLS
            
            callback = partial(toolbar.new_model_item, pos_cm, label, 
                               interactive=interactive)
            
            action.triggered.connect(callback)
            
        
        
        copied_items = self.copied_items
    
        if copied_items is not None:
            ncopied_items = sum([len(items) for key, items in copied_items.items()])                                 
        else:
            ncopied_items = 0
        
        action = context_menu.addAction('Paste item(s) (keep position)')
        callback = lambda: self.paste(pos_cm=None)
        action.triggered.connect(callback)
        action.setEnabled(ncopied_items > 0)
        
        action = context_menu.addAction('Paste item here')
        callback = lambda: self.paste(pos_cm=pos_cm)
        action.triggered.connect(callback)
        action.setEnabled(ncopied_items == 1 and labels.WALLS not in self.copied_items.keys())
        
        
        
        menu_position = QPoint(round(pos.x()), round(pos.y()))
        menu_position = self.view.mapFromScene(menu_position)
        menu_position = self.view.mapToGlobal(menu_position)       
        context_menu.exec_(menu_position)
            
    def add_item(self, label, position):
        self.mpl_controller.view.toolbar.button_checked(label, True)
        self.mpl_controller.add_model_item(label, position)                
  
                    
    def get_snap_action(self, marker):
        
        wall, index, _ = self.model.walls.closestVertexToPixelPoint(
            marker.pos(), exclude_wall=marker.parentItem().model)
        

        closest_vertex = wall.vertices[index]

        disp_vv = str([round(vi) for vi in closest_vertex])
    
        shielding = self.model.shieldings.itemByName(wall.shielding)
        
        action = "Snap to: " + shielding.name + ' at ' + disp_vv
        snap_action = QAction(action)
    
    
        vertex_index = marker.parentItem().markers().index(marker)
        
        
        callback = partial(self.snap_wall, marker.parentItem(), 
                           vertex_index, closest_vertex)
                                   
        snap_action.triggered.connect(callback)
    
        icon = qta.icon('fa5s.circle', color=shielding.color)
        snap_action.setIcon(icon)

        # else:
        #     snap_action = QAction('Snap to: ')
        #     snap_action.setEnabled(False)
        return snap_action
    
    def selection_to_dict(self):
        return [item.to_dict() for item in self.graphics.scene().selectedItems()]
    
    def paste(self, pos_cm=None):
        if self.copied_items is None or len(self.copied_items) == 0:
            return
        
        self.parent.graphics().scene().deselectAll()
        
   
        for label, items in self.copied_items.items():
            sequence = self.model.get_attr_from_label(label) 
            
            for item in items:
                if pos_cm is not None:
                    item.position = pos_cm
                sequence.addItem(item)
                pixel_item = self.parent.graphics().pixel_item_for_model_item(item)
                pixel_item.setSelected(True)
        
        self.parent.graphics().scene().updateStyles()
        

            
        
    def centroid(self, items):
        points = []
        for item in items:
            if isinstance(item, Wall):
                points += [*item.vertices]
            else:
                points += [item.position]
            
        points = np.asarray(points)
        return np.mean(points, 0)
            
        
    
    def show_pick_context_menu(self, pos=None, pixel_item=None):
            pixel_item.setSelected(True)
        
        
            context_menu = QMenu(self.view)
      
                
            if isinstance(pixel_item, WallMarker):
                poly = pixel_item.parentItem()
                vertex_index = poly.markers().index(pixel_item)
                          
                wall = pixel_item.parentItem().model
                index = pixel_item.parentItem().markers().index(pixel_item)
                if not wall.isClosed() and (index == 0 or index == len(wall)-1):
                    text = "Add new marker"
                else:
                    text = "Insert new marker"
            
            
                wall_draw_action = context_menu.addAction(text)            
                callback = partial(self.continue_wall, poly, vertex_index)                        
                wall_draw_action.triggered.connect(callback)
                
                if poly.model.isClosable():
                    close_action = context_menu.addAction("Close polygon")
                    callback = pixel_item.parentItem().model.close
                    close_action.triggered.connect(callback)
                elif poly.model.isClosed():
                    open_action = context_menu.addAction("Open polygon")
                    callback = lambda: pixel_item.parentItem().model.open(vertex_index)
                    open_action.triggered.connect(callback)
                if len(poly.model) > 2:
                    delete_action = context_menu.addAction('Delete marker')
                    callback = lambda: pixel_item.parentItem().model.deleteVertex(vertex_index)
                    delete_action.triggered.connect(callback)
        
                snap_action = self.get_snap_action(marker=pixel_item)
                                                
                                               
                context_menu.addAction(snap_action)
            if pixel_item.isEnabled():
                cut_action = context_menu.addAction("Cut selection")
                cut_action.triggered.connect(self.cut_selected_items)
                     
                copy_action = context_menu.addAction("Copy selection") 
                copy_action.triggered.connect(self.copy_selected_items)
             
                delete_action = context_menu.addAction("Delete selection")
                delete_action.triggered.connect(self.delete_selected_items)
            
            if isinstance(pixel_item, (GraphicsPoint, Cross)):
                enabled_action = context_menu.addAction("Enabled")
                enabled_action.setCheckable(True)            
                enabled_action.setChecked(pixel_item.model.enabled)            
                enabled_action.triggered.connect(pixel_item.model.enable)
             
                
            context_menu.exec_(QCursor.pos())
        
    def delete(self, pixel_item):
        self.model.delete_item(pixel_item.model)
        

    def continue_wall(self, pixel_item, vertex_index):
        vertex = pixel_item.model.positionPixels()[vertex_index]

        if vertex_index == len(pixel_item.model) - 1:
            vertex_index += 1   
        
        pixel_item.model.insertVertexPixelPoint(vertex_index, vertex)
       
        marker = pixel_item.markers()[vertex_index]
        self.parent.toolbar_controller.item_to_move = marker
        
        
    def snap_wall(self, wall, vertex_index, closest_vertex):     
        wall.model.setVertex(vertex_index, closest_vertex)
       
    def copy_selected_items(self):
        selected = self.parent.graphics().scene().selectedGraphicsModelItems()
        
        selected_model_items = [item.model for item in selected]
        
        selected_model_items = self.model.sort_model_items(selected_model_items)
        
        # use deepcopy to prevent anchors and aliases in yaml
        to_yaml = {label: [deepcopy(item.to_dict()) for item in seq]\
                             for label, seq in selected_model_items.items()}
        
        # remove empty walls or sources
        to_yaml = {k: v for k, v in to_yaml.items() if len(v) > 0}
        
            
        self.copy_yaml = yaml.dump(to_yaml, default_flow_style=None)
        
        pyperclip.copy(self.copy_yaml)
        
    
  
    def delete_selected_items(self):        
        selected = self.parent.graphics().scene().selectedGraphicsModelItems()
        self.model.delete_items([item.model for item in selected])
       
        
    def cut_selected_items(self):

        self.copy_selected_items()
        self.delete_selected_items()
        
  


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 19 15:25:31 2023

@author: r757021
"""
import numpy as np

from copy import copy
import math
from matplotlib.colors import LogNorm
from matplotlib.cm import get_cmap
from skimage.measure import find_contours

from PyQt5.QtWidgets import  (QGraphicsEllipseItem, QGraphicsPathItem,  
                              QGraphicsPolygonItem, QGraphicsTextItem,
                              QGraphicsRectItem, QGraphicsPixmapItem)
                              

from PyQt5.QtGui import (QPen, QPainterPath, QPolygonF, QBrush, QColor, 
                         QPixmap, QImage, QFont)
from PyQt5.QtCore import Qt, QPointF,  QRectF,  QSizeF, QObject, pyqtSignal



from pyrateshield import labels
from pyrateshield.gui.legend import Legend

LINE_WIDTH = 'line_width'
LINE_COLOR = 'line_color'
FACE_COLOR = 'face_color'
SIZE       = 'size'
ZVALUE     = 'zvalue'
WIDTH      = 'width'
HEIGHT     = 'height'

def numpy_array_to_qimage(numpy_array):
   
    if len(numpy_array.shape) == 3:
        height, width, channel = numpy_array.shape
    else:
        height, width = numpy_array.shape
        channel = 1
  
    # Create a QImage from the numpy array
    if channel == 3:
        qformat = QImage.Format_RGB888
    elif channel == 4:
        qformat= QImage.Format_RGBA8888
    elif channel == 1:
        qformat = QImage.Format_Grayscale8
        
    qimage = QImage(numpy_array.data, 
                    width, height, width * channel, qformat)
    return qimage


class DosemapPixmap(QGraphicsPixmapItem):
    showContextMenu = False
    _floorplan_size = None
    _dosemap = None
    ZVALUE = -1
    _legend = None
    
    def __init__(self, model=None, 
                       style=None):
                       
        
        super().__init__()
        if style is None:
            raise ValueError
        self.setZValue(self.ZVALUE)   
        self.style = style             
        self.set_callbacks()
        
       
        

        
    def setDosemap(self, dosemap):
        self._dosemap = dosemap
        self.updatePixmap()
        
    def dosemap(self):
        return self._dosemap
        
    def legend(self):
        if self._legend is None:
            self._legend = Legend(model=self.style)
        return self._legend
  
    def set_callbacks(self):        
        self.style.update_event.connect(self.style_update)
   
    def style_update(self, event_data):
        # label, src, old_value, new_value = event_data
 
        self._legend = None
        
        self.updatePixmap()
        
        
    def floorplanSize(self):
        return self.scene().floorplan().pixmap().size()
            
    def coloredDosemap(self):
        scale = LogNorm(self.style.vmin, self.style.vmax, clip=True)
        dosemap = np.flipud(self.dosemap())
        log_dosemap = scale(dosemap)
        log_dosemap = np.ma.getdata(log_dosemap)
        
        cmap = self.getCmap()
        color_dosemap = (cmap(log_dosemap) * 255)
        color_dosemap = color_dosemap.astype(np.uint8)
        return color_dosemap
            
    def updatePixmap(self):
        self.scene().removeContours()
        
        if self.dosemap() is None:
            pixmap = QPixmap()            
        else:
            dosemap_color = self.coloredDosemap()
            qdosemap = numpy_array_to_qimage(dosemap_color)
        
            pixmap = QPixmap.fromImage(qdosemap)
            pixmap = pixmap.scaled(self.floorplanSize())
            self.add_contours()
        
        self.setPixmap(pixmap)
            
    
    def getCmap(self):
        cmap = copy(get_cmap(self.style.cmap_name))        
        alphas = np.linspace(0, 1, cmap.N)
        cmap._init()
        cmap._lut[:len(alphas), -1] = alphas            
        cmap.set_under(alpha=0)
        return cmap

    def add_contours(self): 
        if self.scene() is None: return
        if self.dosemap() is None: return

        self.scene().removeContours()
        
        dosemap = np.flipud(self.dosemap())
        
        
        styles = self.style.contour_lines
        
        if not isinstance(styles, list) or len(styles) == 0:
            raise
        alpha = round(self.style.alpha * 255)
        alpha = min(alpha, 255)
        alpha = max(alpha, 0)
        for style in styles:
            level, line_color, line_style, line_width, enabled = style
            if not enabled: continue
            contour = GraphicsContourItem(dosemap,
                                          level=level,
                                          line_style=line_style,
                                          line_color=line_color,
                                          line_width=line_width,                                          
                                          floorplan_size=self.floorplanSize())
     
            self.scene().addItem(contour)

            
            


class FloorplanPixmap(QGraphicsPixmapItem):
    showContextMenu = True
    ZVALUE = -2
    def __init__(self, *args, model=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.setZValue(self.ZVALUE)
        self.set_callbacks()
        self.updatePixmap()
        
    def set_callbacks(self):
        self.model.update_event.connect(self.model_update)
    
    def model_update(self, event_data):
        label, src, old_value, new_value = event_data
        if label == labels.IMAGE:
            self.updatePixmap()
   
    def updatePixmap(self):
        if self.model is None: return
        qimage = numpy_array_to_qimage(self.model.image)
        self.setPixmap(QPixmap.fromImage(qimage))
    
     
        


class ZoomRect(QGraphicsPolygonItem):
    __pen = None
    def __init__(self, point):
        self.rect = QRectF(point, QSizeF(0,0))
        
        super().__init__(self.poly())
        self.setPen(self._pen())
        
    def poly(self):
        return QPolygonF([self.rect.topLeft(),
                          self.rect.topRight(),
                          self.rect.bottomRight(),
                          self.rect.bottomLeft()])
    
    def _pen(self):
        if self.__pen is None:
    
            self.__pen = QPen()            
            self.__pen.setColor(QColor(0,0,0))
            self.__pen.setWidth(0) # ensures width is always 1 pixel on screen
            self.__pen.setStyle(Qt.DotLine)
            
            
        return self.__pen
        
    def setPoint(self, point):        
        self.rect.setBottomRight(point)
        self.setPolygon(self.poly())
        


class GraphicsItem(QObject):
    showContextMenu         = False
    _highlighted            = False
    _pixel_size_cm          = 1
    
    _style                  = {}
    _highlight_style        = {}
    _select_style           = {}
    _disabled_style         = {}
    
    _highlightable          = True
    _selectable             = True
    _movable                = True
    _send_position_change   = True
    _accept_hover           = True
    _ignore_transformations = True
    
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
      
        self.setFlag(self.ItemIsSelectable, self._selectable)
        self.setFlag(self.ItemIsMovable, self._movable)
        self.setFlag(self.ItemSendsScenePositionChanges, self._send_position_change)
        self.setAcceptHoverEvents(self._accept_hover)
        self.setFlag(self.ItemIgnoresTransformations, self._ignore_transformations)
        self.updateStyle()
       
    def zoomFactor(self):
        try:
            return self.scene().views()[0].zoomFactor()
        except:
            return 1
        
    def ctrlPressed(self):
        if self.scene() is not None:
            return self.scene().views()[0].ctrlPressed()
        else:
            return False
        
    def isHighlighted(self):
        return self._highlighted
    
    def setHighlighted(self, highlighted):
        if self._highlighted != highlighted and self._highlightable:
            self._highlighted = highlighted
            self.updateStyle()
            
    def hoverEnterEvent(self, event):
        if self._highlightable:
            event.accept()
            self.setHighlighted(True)
            
       
    
    def hoverLeaveEvent(self, event):
        if self._highlightable:
            self.setHighlighted(False)
                    
    def pointsToSceneUnits(self, points):
        if self.scene() is None:
            return points
        else:
            return self.scene().pointsToSceneUnits(points)
            
    def itemChange(self, change, value):
        
        if change == self.ItemSelectedHasChanged:    
            self.updateStyle()
        if change == self.ItemSceneHasChanged:
            self.updateStyle()
        return value
    
    def style_zvalue(self):
        return self.style().get(ZVALUE, 0)
    
    def size(self):
        size = self.pointsToSceneUnits(self.style().get(SIZE, 10))
        return size
    
    def width(self):
        size = self.pointsToSceneUnits(self.style().get(WIDTH, 10))
        return size
    
    def height(self):
        size = self.pointsToSceneUnits(self.style().get(HEIGHT, 10))
        return size
        
    
    def linewidth(self):
        width = self.pointsToSceneUnits(self.style().get(LINE_WIDTH, 0)) 
        return width 
    
    def linecolor(self):
        color = self.style().get(LINE_COLOR, [0, 0, 0])        
        color = QColor(*color)
        return color
    
    def facecolor(self):
        color = self.style().get(FACE_COLOR, [0, 0, 0])
        color = QColor(*color)
        return color
        
    def style(self):
        if not self.isEnabled():
            return self._disabled_style
        if self.isSelected():
            return self._select_style
        elif self.isHighlighted() and self._highlightable:
            return self._highlight_style
        else:
            return self._style
                                           
    def _pen(self):
        pen = QPen()

        pen.setColor(self.linecolor())
        pen.setWidthF(self.linewidth())
        return pen
           
    def _brush(self):
        return QBrush(self.facecolor())
        
    def updateStyle(self):
        self.setZValue(self.style_zvalue())
        
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.ctrl_pressed = True
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.ctrl_pressed = False
        super().keyReleaseEvent(event)
        
    def mousePressEvent(self, event):
        self.mouse_press_pos = self.mapToScene(event.pos())
        
    
    
        
        
                        
class GraphicsModelItem(GraphicsItem):
    POSITION_LABEL = labels.POSITION
    _block_updates_from_model  = False
    
    def __init__(self, model=None, pixel_size_cm=None, **kwargs):                
        self.model = model
        self.set_callbacks()
        self._pixel_size_cm = pixel_size_cm
        super().__init__(**kwargs)
        self.modelPositionToGraphics()
        self.setSelected(False)
        
        
    def hoverEnterEvent(self, event):
        super().hoverEnterEvent(event)
        
        if isinstance(self, Marker):
            self.setToolTip(self.parentItem().model.tooltip())
        else:
            self.setToolTip(self.model.tooltip())
            
    def set_callbacks(self):
      
        self.model.update_event.connect(self.modelUpdate)
        
    def itemChange(self, change, value):          
        if change == self.ItemPositionHasChanged:               
            self.graphicsPositionToModel()      
        return super().itemChange(change, value)                                       

    def graphicsPositionToModel(self):
        self._block_updates_from_model = True
        self.model.setPositionPixels(self.pos())
        self._block_updates_from_model = False
    
    def modelPositionToGraphics(self):
        self.setPos(self.model.positionPixels())

    def modelUpdate(self, event_data):
        src, label, old_value, value = event_data   
        if label == self.POSITION_LABEL:
            if not self._block_updates_from_model:
                self.modelPositionToGraphics()
        elif label == labels.ENABLED:
            self.setEnabled(value)
            self.updateStyle()
     
            
class GraphicsPoint(QGraphicsEllipseItem, GraphicsModelItem):  
    showContextMenu         = True
    POSITION_LABEL = labels.POSITION
    mouse_press_pos = QPointF()
    __pen = None
    __brush = None
    
    _style = {SIZE: 20,
              FACE_COLOR: [0, 255, 0, 128],
              LINE_COLOR: [0, 0, 0],
              LINE_WIDTH: 1,
              ZVALUE: 5}
    
    _highlight_style = {SIZE: 50,
                  FACE_COLOR: [0, 255, 0, 128],
                  LINE_COLOR: [0, 0, 0],
                  LINE_WIDTH: 1,
                  ZVALUE: 10}
    
    _select_style = {SIZE: 50,
                  FACE_COLOR: [0, 255, 0, 128],
                  LINE_COLOR: [0, 0, 0],
                  LINE_WIDTH: 1,
                  ZVALUE: 20}
    
    _disabled_style = {SIZE: 10,
                       FACE_COLOR: [83, 83, 83],
                       LINE_COLOR: [0, 0, 0],
                       LINE_WIDTH:2,
                       ZVALUE: 5}
    
    
    
    def __init__(self, model=None, pixel_size_cm=1, **kwargs):
        super().__init__(model=model, pixel_size_cm=pixel_size_cm, **kwargs)
       
        

    
    

    def drawRect(self):
        return QRectF(QPointF(-self.radius(), -self.radius()),
                      QSizeF(2*self.radius(), 2*self.radius()))
    
    def radius(self):
        return  self.size() / 2

    def updateStyle(self):
        super().updateStyle()
        self.setPen(self._pen())
        self.setBrush(self._brush())
        self.setRect(self.drawRect())
        
    def mousePressEvent(self, event):
        self.mouse_press_pos = self.pos()
        
    def itemChange(self, change, value):
        if change == self.ItemPositionChange and self.ctrlPressed():
            if self.scene() is not None and len(self.scene().selectedItems()) == 1:
                if self.mouse_press_pos is not None:
                    reference_pos = self.mouse_press_pos
                
                    dx = abs(reference_pos.x() - value.x())
                    dy = abs(reference_pos.y() - value.y())
                    
                    if dx >= dy:
                        value = QPointF(value.x(), reference_pos.y())
                    else:
                        value = QPointF(reference_pos.x(), value.y())
                
        return GraphicsModelItem.itemChange(self, change, value)  
    #     
    
    

class GraphicsSourceNM(GraphicsPoint):
    _style = {SIZE: 10,
              FACE_COLOR: [139, 0, 0, 128],
              LINE_COLOR: [0, 0, 0],
              LINE_WIDTH: 2,
              ZVALUE: 5}
    
    _highlight_style = {SIZE: 20,
                  FACE_COLOR: [139, 0, 0, 128],
                  LINE_COLOR: [0, 0, 0],
                  LINE_WIDTH: 2,
                  ZVALUE: 10}
    
    _select_style = {SIZE: 20,
                  FACE_COLOR: [139, 0, 0, 128],
                  LINE_COLOR: [0, 0, 0],
                  LINE_WIDTH: 2,
                  ZVALUE: 20}
    

class GraphicsSourceCT(GraphicsPoint):
    _style = {SIZE: 10,
              FACE_COLOR: [34, 139, 35, 128],
              LINE_COLOR: [0, 0, 0],
              LINE_WIDTH: 2,
              ZVALUE: 5}
    
    _highlight_style = {SIZE: 20,
                  FACE_COLOR: [34, 139, 35, 128],
                  LINE_COLOR: [0, 0, 0],
                  LINE_WIDTH: 2,
                  ZVALUE: 10}
                  
    _select_style = {SIZE: 20,
                     FACE_COLOR: [34, 139, 35, 128],
                     LINE_COLOR: [0, 0, 0],
                     LINE_WIDTH: 2,
                     ZVALUE: 20}
    

class GraphicsSourceXray(GraphicsPoint):
    _style = {SIZE: 10,
              FACE_COLOR: [34, 34, 139, 128],
              LINE_COLOR: [0, 0, 0],
              LINE_WIDTH: 2,
              ZVALUE: 5}
    
    _highlight_style = {SIZE: 20,
                        FACE_COLOR: [34, 34, 139, 128],
                        LINE_COLOR: [0, 0, 0],
                        LINE_WIDTH: 2,
                        ZVALUE: 10}
    
    _select_style = {SIZE: 20,
                     FACE_COLOR: [34, 34, 139, 128],
                     LINE_COLOR: [0, 0, 0],
                      LINE_WIDTH: 2,
                      ZVALUE: 20}
    

    


class Cross(QGraphicsPolygonItem, GraphicsModelItem):
    showContextMenu         = True
    _style = {
           
              LINE_COLOR: [238, 75, 43, 128],
              LINE_WIDTH: 3,
              WIDTH: 15,
              HEIGHT: 15,
              ZVALUE: 5}
              
    
    _highlight_style = {
                  
                  LINE_COLOR: [238, 75, 43, 128],
                  LINE_WIDTH: 6,
                  WIDTH: 30,
                  HEIGHT: 30,              
                  ZVALUE: 10}
                 
    
    _select_style = {
                  
                  LINE_COLOR: [238, 75, 43, 128],
                  LINE_WIDTH: 6,
                  WIDTH: 30,
                  HEIGHT: 30,
                  ZVALUE: 20}
    
    _disabled_style = {LINE_COLOR: [83, 83, 83],
                       LINE_WIDTH: 3,
                       WIDTH: 15,
                       HEIGHT: 15,
                       ZVALUE: 5}
                  
    
    
    def __init__(self, model=None, **kwargs):

        super().__init__(model=model, **kwargs)
        self.modelPositionToGraphics()
        self.updateStyle()
       
    
    def points(self):
        return [[-1, -1],
                [ 1, 1],
                [0, 0],
                [-1, 1],
                [1, -1], 
                [0, 0]]
  
    
    
    def _polygon(self):
        w, h = self.width(), self.height()
        
        Xp = [QPointF(xi[0]*w*0.5, xi[1]*h*0.5) for xi in self.points()]
        
        return QPolygonF(Xp)
    
    def updateStyle(self):
        super().updateStyle()
        self.setPen(self._pen())
        self.setPolygon(self._polygon())
        
    def itemChange(self, change, value):
        if change == self.ItemPositionChange and self.ctrlPressed():
            if self.scene() is not None and len(self.scene().selectedItems()) == 1:
                if self.mouse_press_pos is not None:
                    reference_pos = self.mouse_press_pos
                
                    dx = abs(reference_pos.x() - value.x())
                    dy = abs(reference_pos.y() - value.y())
                    
                    if dx >= dy:
                        value = QPointF(value.x(), reference_pos.y())
                    else:
                        value = QPointF(reference_pos.x(), value.y())
                
        return GraphicsModelItem.itemChange(self, change, value) 
        

class Marker(QGraphicsRectItem, GraphicsItem):
    showContextMenu         = True
    _style = {WIDTH: 10,
              HEIGHT: 10,
              LINE_WIDTH: 2,
              ZVALUE: 3}
         
    _highlight_style = {WIDTH: 20,
                        HEIGHT: 20,
                        LINE_WIDTH: 2,
                        ZVALUE: 10}
    
    _select_style = {WIDTH: 20,
                     HEIGHT: 20,
                     LINE_WIDTH: 2,
                     ZVALUE: 20}
    
    def __init__(self, **kwargs):
        #self.line = line
        
        super().__init__(**kwargs)
        
    def style_zvalue(self):
        if any([marker.isSelected() for marker in self.otherMarkers()]):
            return self._select_style[ZVALUE]
        else:
            return super().style_zvalue()
            

    def _pen(self):
        return QPen(Qt.NoPen)
        
    def otherMarkers(self):
        return [marker for marker in self.parentItem().markers()\
                if marker is not self]
            
    def adjecentMarker(self):
        index = self.parentItem().markers().index(self)
        if index > 0:
            adjecent = self.parentItem().markers()[index-1]
        else:
            adjecent = self.parentItem().markers()[1]
        return adjecent
            
            
    def index(self):
        return self.parentItem().markers().index(self)
    
    def isHighlighted(self):
        return self._highlighted or any([marker._highlighted for marker in self.otherMarkers()])
               
    def _rect(self):
        return QRectF(-0.5*self.width(), -0.5*self.height(), 
                      self.width(), self.height())
    
    def facecolor(self):
        if self.parentItem() is None: return QColor()
        color = self.parentItem().linecolor()    
        color.setAlpha(128)
        return color
    
    def linecolor(self):
        if self.parentItem() is None: return QColor()
        return self.parentItem().linecolor()
    
    def itemChange(self, change, value):        
        if change == self.ItemSelectedHasChanged: 
            if not self.isSelected():
                self.setHighlighted(False)
            self.updateStyle()
        elif change == self.ItemPositionHasChanged:
            self.parentItem().graphicsPositionToModel()
            self.parentItem().modelPositionToGraphics()
        
        elif change == self.ItemPositionChange and self.ctrlPressed():
            if self.scene() is not None and len(self.scene().selectedItems()) == 1:

                reference_pos = self.adjecentMarker().pos()
                
            
                dx = abs(reference_pos.x() - value.x())
                dy = abs(reference_pos.y() - value.y())
                
                if dx >= dy:
                    value = QPointF(value.x(), reference_pos.y())
                else:
                    value = QPointF(reference_pos.x(), value.y())
                        
            return super().itemChange(change, value)  
            
       
        
    

        return super().itemChange(change, value)
    
    def updateStyle(self, update_others=True):    
        super().updateStyle()        
        self.setBrush(self._brush())
        self.setRect(self._rect())    
        self.setPen(self._pen())        
        if update_others:
            for marker in self.otherMarkers():
                marker.updateStyle(update_others=False)
                
   
class WallMarker(Marker):
    pass
        

   
class GraphicsLine(QGraphicsPathItem, GraphicsModelItem):
    POSITION_LABEL = labels.VERTICES
    markerClass = Marker
    _style = {LINE_WIDTH: 2,
              LINE_COLOR:  [0, 0, 139],
              ZVALUE: 0}
    
    # _highlight_style = {LINE_COLOR:  [0, 255, 0],                        
    #                     LINE_WIDTH: 0}
    
    
    # _select_style = {LINE_COLOR:  [0, 0, 139],                     
    #                  LINE_WIDTH: 0,
    #                  ZVALUE: 0}
    
    _pixel_size_cm = 1
    _markers = None
    _polygon = None
    
    _selectable = False
    _movable = False
    _highlightable = False
    _ignore_transformations = False
    
    
    def markerPolygon(self):
        return QPolygonF([marker.pos() for marker in self.markers()])
    
    def graphicsPositionToModel(self):          
            self._block_updates_from_model = True
            self.model.setPositionPixels(self.markerPolygon())
            self._block_updates_from_model = False
            
    def setSelected(self, selected):
        for marker in self.markers():
            marker.setSelected(selected)
            
    def isSelected(self):
        return all([marker.isSelected() for marker in self.markers()])
        
    def modelPositionToGraphics(self):
        poly = self.model.positionPixels()
        if self.model.isClosed():
            poly.append(poly[0])
        self.setPolygon(poly)
        self.updateMarkerPosition()
    
    def polygon(self):
        if self._polygon is None:
            self._polygon = QPolygonF()
        return self._polygon
    
    def setPolygon(self, polygon):
        self._polygon = polygon
        path = QPainterPath()
        path.addPolygon(polygon)
        self.setPath(path)
    
    def markers(self):
        if self._markers is None:
            self._markers = []
        return self._markers
    
    def updateMarkerPosition(self):
        all_selected = all([marker.isSelected() for marker in self.markers()])
        new_markers = len(self.model.vertices) - len(self.markers())
        markers = [self.markerClass(parent=self) for i in range(0, new_markers)]
        self.markers().extend(markers)
        
        if all_selected:
            for marker in markers:
                marker.setSelected(True)
               
        #delete_markers = len(self.markers()) - len(self.model.vertices)
        
        while len(self.markers()) > len(self.model.vertices):
            marker = self.markers().pop(0)
            self.scene().removeItem(marker)
            marker.setParentItem(None)
            del marker
            
        for point, marker in zip(self.polygon(), self.markers()):
            marker.setPos(point)
    
    def updateLineStyle(self):       
        self.setPen(self._pen())
    
    
    def modelUpdate(self, event_data):
        super().modelUpdate(event_data)
        src, label, old_value, value = event_data
        if label == labels.CLOSED:
           self.setSelected(False)
                
    def itemChange(self, change, value):
        if change == self.ItemSceneHasChanged:                        
            self.updateStyle()
        
        return super().itemChange(change, value)
            
    def updateStyle(self):
        super().updateStyle()
        #self.updateMarkers()
        self.setPen(self._pen())   
        for marker in self.markers():
            marker.updateStyle()
            
    def linewidth(self):
        width = self.pointsToSceneUnits(self.style().get(LINE_WIDTH, 0)) / self.zoomFactor()
        return width 

    
        
class GraphicsWall(GraphicsLine):   
    markerClass = WallMarker
    def __init__(self, model=None, pixel_size_cm=1):
        self.model = model
        self.shieldings().styleChanged.connect(self.updateStyle)
        self.model.parent().shieldingChanged.connect(self.updateStyle)
        super().__init__(model=model, pixel_size_cm=pixel_size_cm)
               
        
        
    def _pen(self):
        pen = super()._pen()
        if self.shielding().name == labels.EMPTY_SHIELDING:
            pen.setStyle(Qt.DotLine)
        return pen
    
    def modelUpdate(self, event_data):
        super().modelUpdate(event_data)
        src, label, old_value, value = event_data     
        if label == labels.CLOSED:
            if not self._block_updates_from_model:
                self.modelPositionToGraphics()
        
        
    def linecolor(self):
        color = self.shielding().color
        if isinstance(color, str):
            color = QColor(self.shielding().color)
        elif isinstance(color, (tuple, list)):
            color =  QColor(*self.shielding().color)
        return color
    
    def linewidth(self):
        width = self.pointsToSceneUnits(self.shielding().linewidth) / self.zoomFactor()
        
        return width
        
    def shieldings(self):             
        return self.model.parent().parent().shieldings
    
    def shielding(self):       
        for item in self.shieldings().items:                      
            if item.name == self.model.shielding:
                return item
        return self.shieldings()[0]
    
    def mousePressEvent(self, event):
        super().mouseMoveEvent(event)
      
        if self.scene() is not None:
            self.scene().deselectAll()
    
    
    
          
class PixelSizeMarker(Marker):
    
    _style = {SIZE: 10,
              LINE_WIDTH: 2,
              ZVALUE: 3}
         
    _highlight_style = {SIZE: 20,
                        LINE_WIDTH: 2,
                        ZVALUE: 10}
    
    _select_style = _style
    
    _highlightable = False
    
    def facecolor(self):
        color = self.parentItem().linecolor()    
        color.setAlpha(64)
        return color
    
    def _pen(self):
        pen = QPen()
        pen.setColor(self.parentItem().linecolor())
        pen.setWidth(0)
        pen.setStyle(Qt.DotLine)
        return pen
    
class RulerMarker(Marker):
    showContextMenu = False
    _style = {WIDTH: 6,
              HEIGHT: 20,
              LINE_WIDTH: 0,
              ZVALUE: 3}
         
    _highlight_style = {WIDTH: 6,
                        HEIGHT: 20,
                        LINE_WIDTH: 0,
                        ZVALUE: 10}
    
    _select_style = {WIDTH: 6,
                     HEIGHT: 20,
                     LINE_WIDTH: 0,
                     ZVALUE: 20}
    
    def facecolor(self):
        color = self.parentItem().linecolor()    
        color.setAlpha(255)
        return color
    
    def _pen(self):
        pen = QPen()
        pen.setColor(self.parentItem().linecolor())
        pen.setWidth(1)
        pen.setStyle(Qt.SolidLine)
        return pen
    
    
    def _rect(self):
         if self.parentItem() is None or self not in self.parentItem().markers():
             return QRectF(-0.5*self.width(), -0.5*self.height(), 
                           self.width(), self.height())
         
         index = self.parentItem().markers().index(self)
         lw = self.pointsToSceneUnits(self.parentItem()._style[LINE_WIDTH])
         
         
         if index == 1:
             return QRectF(-self.width()+0.5*lw, -0.5*self.height(), 
                           self.width(), self.height())
         if index == 0:
             return QRectF(-0.5*lw, -0.5*self.height(), 
                           self.width(), self.height())
    
    
# class PixelSizeLine(GraphicsLine):    
#     POSITION_LABEL = labels.VERTICES_PIXELS
#     markerClass = PixelSizeMarker
#     #_pixel_size_cm = 1    
    
    
    
#     def graphicsPositionToModel(self):
#         vertices_pixels = [[marker.pos().x(), marker.pos().y()]\
#                            for marker in self.markers()]
        
#         self._block_updates_from_model = True
#         self.model.vertices_pixels = vertices_pixels
#         self._block_updates_from_model = False
        
#     def modelPositionToGraphics(self):       
#         p1 = QPointF(*self.model.vertices_pixels[0])
#         p2 = QPointF(*self.model.vertices_pixels[1])
#         self.setPolygon(QPolygonF((p1, p2)))
#         self.updateMarkerPosition()
        
        
        
class RulerLine(GraphicsLine):
    POSITION_LABEL = labels.VERTICES_PIXELS
    showContextMenu = False
    markerClass = RulerMarker
    TEXT_SIZE = 10
    def __init__(self, *args, show_text=True, **kwargs):
        self.show_text = show_text
        if self.show_text:
            self.text = QGraphicsTextItem("test")
       
        super().__init__(*args, **kwargs)
        if self.show_text:
            self.text.setParentItem(self)
            
            self.text.setFlag(self.ItemIgnoresTransformations, True)
        
    def _pen(self):
        pen = super()._pen()
        pen.setStyle(Qt.DashLine)
        return pen
    
    def font(self):
        font = QFont()
        
        size = self.pointsToSceneUnits(self.TEXT_SIZE)
        font.setPointSizeF(size)
        return font
    
    def modelPositionToGraphics(self):     
        p1 = QPointF(*self.model.vertices_pixels[0])
        p2 = QPointF(*self.model.vertices_pixels[1])
        self.setPolygon(QPolygonF((p1, p2)))
        self.updateMarkerPosition()
        
        self.markers()[0].setRotation(self.model.angle)
        self.markers()[1].setRotation(self.model.angle)
        if self.show_text:
            self.updateText()
            
    def graphicsPositionToModel(self):
         vertices_pixels = [[marker.pos().x(), marker.pos().y()]\
                            for marker in self.markers()]
         
         self._block_updates_from_model = True
         self.model.vertices_pixels = vertices_pixels
         self._block_updates_from_model = False
        
    def updateStyle(self):
        super().updateStyle()
        if self.show_text:
            self.updateText()
    
    def updateText(self):
        length = round(self.model.length_cm, 1)
        self.text.setPlainText(f'{length} cm')
        poly = self.polygon()
        if len(poly) == 0: return
        p1 = poly[0]
        p2 = poly[1]
        dy = p2.y() - p1.y()
        dx =  p2.x() - p1.x()
        line_angle = math.atan2(dy, dx)
        
        text_width = self.text.boundingRect().width()
        
        center = [(p2.x() + p1.x())/2, (p2.y() + p1.y())/2]
        
        #x2 + y2 = tw2
        
        #xi = center[0] + text_width/2 * math.sin(line_angle)
        #yi = center[1] + text_width/2 * math.cos(line_angle)
        
        xi = center[0] 
        yi = center[1]
        
        #self.text.setRotation(math.degrees(line_angle))
        self.text.setPos(QPointF(xi, yi))
        self.text.setFont(self.font())
        
        
    
    
class GraphicsContourItem(QGraphicsPathItem, GraphicsItem):
    _selectable = False
    _highlightable = False
    _movable = False
    _ignore_transformations = False
    
    def __init__(self, dosemap, level=1, line_style='solid', line_color='red',
                 line_width=0, floorplan_size=None):
        
        
        self.dosemap = dosemap
        self.level = level
        self.floorplan_size = floorplan_size
        self._line_style = line_style
        self._line_width = line_width        
        self._line_color = line_color
        super().__init__()
        self.setPath(self.path())
        self.updateStyle()
                      
    def updateStyle(self):
        self.setPen(self._pen())
        
    def linecolor(self):
        if isinstance(self._line_color, (list, tuple)):
            color = QColor(*self._line_color)
        else:
            color = QColor(self._line_color)                  
        return color
    
    def hoverEnterEvent(self, event):
        super().hoverEnterEvent(event)
        self.setToolTip(f'Dose level: {self.level} mSv')
            
    
        
    def linestyle(self):
        line_style = self._line_style
        if line_style == 'dashed':
            line_style = Qt.DashLine
        elif line_style == 'solid':
            line_style = Qt.SolidLine
        elif line_style == 'dashdot':
            line_style = Qt.DashDotLine
        elif line_style == 'dotted':
            line_style = Qt.DotLine
        return line_style
    
               
    def lineWidth(self):
        width = self.pointsToSceneUnits(self._line_width) / self.zoomFactor()
   
        return width

    def _pen(self):                    
        pen = QPen(self.linecolor(), self.lineWidth(), self.linestyle())
        return pen
    
    def path(self):
        path = QPainterPath()
        for polygon in self.polygons():            
            path.addPolygon(polygon)           
        return path
        
    def np_contours(self):
        return find_contours(self.dosemap, self.level)
    
    def polygons(self):
        polygons = []
        sx = self.floorplan_size.width()
        sy = self.floorplan_size.height()
        scale_x = sx / self.dosemap.shape[1]
        scale_y = sy / self.dosemap.shape[0]
        
        for np_contour in self.np_contours():
            points = [QPointF(pi[1]*scale_x, pi[0]*scale_y) for pi in np_contour]
            polygon = QPolygonF(points)
            polygons += [polygon]
        return polygons
    
    


import os
import yaml
import qtawesome
import math

from functools import partial
from scipy.ndimage import zoom

from PyQt5.QtWidgets import QToolBar, QMainWindow, QApplication
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QIcon

from pyrateshield import labels
from pyrateshield.gui.graphics_items import ZoomRect,  Marker



def resize_array(source_array, target_shape):
    # Need faster method for painting dosemap on top of the floorplan.
    # Obtain same size by interpolation
    zoom_factors = [float(t) / s for s, t in zip(source_array.shape, target_shape)]
    resized_array = zoom(source_array, zoom_factors, order=1)
    return resized_array




class ToolbarController():
     item_to_move = None
     _zoomRect = None
     def __init__(self, view=None, model=None, parent=None):
         self.parent = parent
         self.view = view
         self.model = model
         self.graphics = self.parent.view.views[labels.CANVAS]
         
         
         self.set_callbacks()
    
     def set_callbacks(self):
         self.view.buttonClick.connect(self.toolbar_cbk)
         self.graphics.leftMouseClick.connect(self.leftClick)
         self.graphics.mouseMove.connect(self.mouseMove)
         self.graphics.mouseRelease.connect(self.mouseRelease)
        
        
     def leftClick(self, event, item):
         scene_pos = self.graphics.mapToScene(event.pos())
         self.click_pos = scene_pos
         if self.view.selected_tool() in (labels.SOURCES_NM,
                                          labels.SOURCES_CT,
                                          labels.SOURCES_XRAY,
                                          labels.CRITICAL_POINTS,
                                          labels.WALLS,
                                          labels.RULERS):
                                          
             
             pos_cm = self.mouse_pos_to_cm(event.pos())
             self.new_model_item(pos_cm, self.view.selected_tool())
             
             
         elif self.view.selected_tool() == self.view.ZOOM:      
                  
            self.item_to_move = ZoomRect(scene_pos)
            self.graphics.scene().addItem(self.item_to_move)
            
         elif self.view.selected_tool() == Toolbar.SELECT:
            self.item_to_move = ZoomRect(scene_pos)
            self.graphics.scene().addItem(self.item_to_move)
            
    
     def mouseMove(self, event):
         scene_pos = self.graphics.mapToScene(event.pos())
         
         
         if self.item_to_move is not None:
             if isinstance(self.item_to_move, Marker):
                 if  event.modifiers() == Qt.ControlModifier:
                     pos_adjecent = self.item_to_move.adjecentMarker().pos()
                     dx = abs(pos_adjecent.x() - scene_pos.x())
                     dy = abs(pos_adjecent.y() - scene_pos.y())
                     if dx >= dy:
                         scene_pos.setY(pos_adjecent.y())
                     else:
                         scene_pos.setX(pos_adjecent.x())
                  
                 self.item_to_move.setPos(scene_pos)
                 
             elif isinstance(self.item_to_move, ZoomRect):                 
                 self.item_to_move.setPoint(scene_pos)
      
     
     def mouseRelease(self, event):      
        self.mouseMove(event)
        if self.view.selected_tool() == Toolbar.ZOOM:
            self.finishZoom()   
        elif self.view.selected_tool() == Toolbar.SELECT:
            self.finishSelect()
            
            
        self.item_to_move = None
        self.graphics.setCursor(Qt.ArrowCursor)
        
     def finishSelect(self):
         selection = self.item_to_move.rect
       
         for item in self.graphics.scene().itemsInRect(selection):
             item.setSelected(True)
         self.graphics.scene().removeItem(self.item_to_move)
         self.view.uncheck_all()          
         
            
     def finishZoom(self):      

        self.graphics.setZoomRect(self.item_to_move.rect)
        self.graphics.scene().removeItem(self.item_to_move)  
        
        self.view.uncheck_all()          
            
         
     def show_dosemap(self, engine):
         
         self.model.dosemap.engine = engine
         
         
         dosemap = self.parent.dosemapper.get_dosemap(self.model)
         
         
         if dosemap is None:
             print('No dosemap could be calculated. Possibly no sources defined or all sources are disabled!')
         else:             
             self.graphics.scene().dosemap().setDosemap(dosemap)
         self.parent.graphics().viewport().update()
         
            
     
         
             
     def delete(self):                
        items = self.graphics.scene().selectedGraphicsModelItems()
        self.model.delete_items([item.model for item in items])
         
        
     def toolbar_cbk(self, item):     
         if item == Toolbar.PYSHIELD:
             self.show_dosemap(labels.PYSHIELD)
         elif item == Toolbar.RADTRACER:
             self.show_dosemap(labels.RADTRACER)
         elif item == Toolbar.RESET:
             self.graphics.resetZoom()
         elif item == labels.FLOORPLAN:
              self.graphics.refresh()
         elif item == Toolbar.DELETE:
             self.delete()
         elif item == Toolbar.ALL:
            self.graphics.scene().selectAll()
         elif item == Toolbar.CLEAR:
            self.graphics.scene().deselectAll()
         elif item == Toolbar.UNDO:
            self.model.history.undo()
         elif item == Toolbar.REDO:
            self.model.history.redo()
         
             
             
     def refresh(self):
        self.graphics.refresh()
        
    
     def mouse_pos_to_cm(self, pos):
         geometry = self.model.floorplan.geometry        
         pos = self.graphics.mapToScene(pos)
         pos_cm = geometry.pixels_to_cm([pos.x(), pos.y()])
         return pos_cm
     
     def new_model_item(self, pos_cm, item_label, interactive=True):
        # copy values of item that are currently in view to new model

        
        sequence = self.model.get_attr_from_label(item_label)
        if item_label == labels.RULERS:
            item_model = sequence.item_class()
        else:
            current_item = self.parent.controllers[item_label].item_in_view()

            if current_item is None:            
                item_model = sequence.item_class()
            else:
                item_model = sequence.item_class.from_dict(current_item.to_dict()) 
            

        if item_label == labels.WALLS:      
            controller = self.parent.controllers[labels.SHIELDINGS]
            item_model.shielding = controller.item_in_view().name
            item_model.vertices = [pos_cm, pos_cm]
        elif item_label == labels.RULERS:
            item_model.vertices = [pos_cm, pos_cm]

        else:            
            item_model.name = item_model.default_name
            item_model.position = pos_cm     
                   
        
        sequence.addItem(item_model)
        
        # Style item as selected        
        pixel_item = self.graphics.pixel_item_for_model_item(item_model)  
        pixel_item.updateStyle()
        self.graphics.scene().deselectAll()
        if item_label in (labels.WALLS, labels.RULERS):
            pixel_item.markers()[1].setSelected(True)
        else:
            pixel_item.setSelected(True)
       
        
        if interactive:                      
            if item_label in (labels.WALLS, labels.RULERS):
                self.item_to_move = pixel_item.markers()[1]
            else:
                pass
        self.view.uncheck_all()
                # self.mouse._point_setter = pixel_item.setCenterPoint    

class Toolbar(QToolBar):
    buttonClick = pyqtSignal(str)
    # only display the buttons we need
    # EVENT_TOOLBUTTON_CLICK = 'event_toolbutton_click'
    _toolbutton_style = Qt.ToolButtonTextUnderIcon
    
    # explicit naming to easily catch events

    NEW = 'New'
    IMAGE = 'Image'
    
    SEPERATOR = 'Seperator'
    RESET = 'Reset'
    
    ZOOM = 'Zoom'
    SELECT = 'Select'
    ALL = 'All'
    CLEAR = 'Clear'
    
    UNDO = 'Undo'
    REDO = 'Redo'
    
    
    DELETE = 'Del'
    SNAPSHOT = 'Snap'
    LOAD = 'Load'
    SAVE = 'Save'
    SAVE_AS = 'Save As'
    CALCULATOR = 'Calc'
    
    DOSEMAP_STYLE = 'Style'    
    ISOTOPES = 'Isotopes'
    FLOORPLAN = 'Floorplan'
    CRITICAL_POINT = 'Point'
    SOURCE_NM = 'NM'
    SOURCE_CT = 'CT'
    SOURCE_XRAY = 'Xray'
    RULER = 'Ruler'
    WALL = 'Wall'
    PYSHIELD = 'pyShield'
    RADTRACER = 'Radtracer'
    
    # Use shorter names to display in GUI when clicked return the name 
    # used throughout pyrateshield
    tool_name_map = {DOSEMAP_STYLE: labels.DOSEMAP_STYLE,
                     ISOTOPES: labels.ISOTOPES,
                     FLOORPLAN: labels.FLOORPLAN,
                     CRITICAL_POINT: labels.CRITICAL_POINTS,
                     SOURCE_NM:  labels.SOURCES_NM,
                     SOURCE_CT: labels.SOURCES_CT,
                     WALL: labels.WALLS,
                     PYSHIELD: labels.PYSHIELD,
                     RADTRACER: labels.RADTRACER,
                     SOURCE_XRAY: labels.SOURCES_XRAY,
                     RULER: labels.RULERS}

    def __init__(self):
        QToolBar.__init__(self)
      
        
        self.setToolButtonStyle(self._toolbutton_style)
        #self.setStyleSheet("border-bottom: 0px; border-top: 0px;")
        self.populate_toolbar()
        
    def get_layout(self):
        try:
            folder = os.path.split(__file__)[0]
        except:
            folder = os.getcwd()
            
        file = os.path.join(folder, 'toolbar.yml')
        return yaml.safe_load(open(file))
        
    def enable_tool(self, label, enable):
        self._actions[label].setEnabled(enable)
    
    
    def populate_toolbar(self):
        layout = self.get_layout()
        self._actions = {}
        self._checkable_tools = []
        for item in layout:
            if item[labels.LABEL] == self.SEPERATOR:
                self.addSeparator()
            else:
                icon = self.icon(item[labels.ICON])
                action = self.addAction(icon, item[labels.LABEL])
                action.setToolTip(item[labels.TOOILTIP])                
                action.setCheckable(item[labels.CHECKABLE])
                if item[labels.CHECKABLE]:
                    self._checkable_tools += [item[labels.LABEL]]
                    cbk = partial(self.checkable_tool_clicked, action)
                    action.triggered.connect(cbk)
                else:
                    cbk = partial(self.emit, item[labels.LABEL])
                    action.triggered.connect(cbk)
                self._actions[item[labels.LABEL]] = action
    
        
    def checkable_tool_clicked(self, clicked_action):
       
        for action in self._actions.values():
            if action is clicked_action:
                continue
            action.setChecked(False)
            
        
                
    def icon(self, name):
        if name == 'icon':
           
            icon = QIcon('../icon.ico')
        else:
            icon = qtawesome.icon(name)
        return icon
    
    def emit(self, tool_name):

        if tool_name in self._checkable_tools:
            self.uncheck_all(except_tool=tool_name)
        else:
            self.uncheck_all()
        
        tool_name = self.tool_name_map.get(tool_name, tool_name)
            
        self.buttonClick.emit(tool_name)
        
   
    def selected_tool(self):
        for tool_name in self._checkable_tools:
            widget = self.widgetForAction(self._actions[tool_name]) 
            if widget.isChecked():
                return self.tool_name_map.get(tool_name, tool_name)
        

    def uncheck_all(self, except_tool=None):        
        for tool_name in self._checkable_tools: 
      
            if tool_name == except_tool:
                continue
            
            action = self._actions[tool_name]
            if action.isChecked():
                action.setChecked(False)

    
    

if __name__ == '__main__':
    app = QApplication([])
    
    
    project = '../example_projects/SmallProject/project.zip'

    
    window = QMainWindow()
    toolbar = Toolbar()
    
    window.addToolBar(toolbar)
    toolbar.actions()[-11].setChecked(True)
    toolbar.actions()[-11].setChecked(False)

    

    
    window.show()
    app.exec_()
    


from pyrateshield.constants import CONSTANTS
from pyrateshield.model import DEFAULTS
from pyrateshield.pyshield.doserates import H10
from pyrateshield.pyshield.isotope import AttenuationHelper, BuildupHelper

def get_material_by_name(name):
    return [material for material in DEFAULTS.materials if material.name==name][0]

def dose_rate(isotope_name):
    isotope = CONSTANTS.get_isotope_by_name(isotope_name)
    
    energy_keV, abundance = list(zip(*isotope.spectrum))
    
    abundance = [ai / 100 for ai in abundance] #% to fraction
    
    return H10(energy_keV, abundance)


def attenuation_isotope(isotope_name, material, thickness):
   
    
    isotope = CONSTANTS.get_isotope_by_name(isotope_name)
    
    energy_keV, abundance = list(zip(*isotope.spectrum))
    
    abundance = [ai / 100 for ai in abundance] #% to fraction
    
    att = [AttenuationHelper.calculate(material, ei,thickness) for ei in energy_keV]
    
    unshielded_dose_rate = H10(energy_keV, abundance)
    
    shielded_abundance = [ai * atti for ai, atti in zip(abundance, att)]
    
    shielded_dose_rate = H10(energy_keV, shielded_abundance)
    
    return shielded_dose_rate / unshielded_dose_rate


def buildup_isotope(isotope_name, material, thickness):
    
    
    isotope = CONSTANTS.get_isotope_by_name(isotope_name)
    
    energy_keV, abundance = list(zip(*isotope.spectrum))

    abundance = [ai / 100 for ai in abundance] #% to fraction
    
    bup = [BuildupHelper.calculate(material, ei, thickness) for ei in energy_keV]
    
    unshielded_dose_rate = H10(energy_keV, abundance)
    
    shielded_abundance = [ai * atti for ai, atti in zip(abundance, bup)]
    
    shielded_dose_rate = H10(energy_keV, shielded_abundance)
    
    return shielded_dose_rate / unshielded_dose_rate


def transmission_isotope(isotope_name, material, thickness):
    
    
    isotope = CONSTANTS.get_isotope_by_name(isotope_name)

    energy_keV, abundance = list(zip(*isotope.spectrum))
    
    abundance = [ai / 100 for ai in abundance] #% to fraction
    
    bup = [BuildupHelper.calculate(material, ei, thickness) for ei in energy_keV]
    
    att = [AttenuationHelper(material, ei, thickness) for ei in energy_keV]
    
    trans = [ai * atti * bupi for ai, atti, bupi in zip(abundance, att, bup)]
    
    shielded_dose_rate = H10(energy_keV, trans)
    
    unshielded_dose_rate = H10(energy_keV, abundance)
    
    return shielded_dose_rate / unshielded_dose_rate


if __name__ == "__main__":
    pass
